#include "Nehebkau.h"
#include <avr/io.h>
#include <avr/interrupt.h>

#define SER_TXBUFFSIZE 16
static volatile uint16_t ser_txBuff[SER_TXBUFFSIZE];
static volatile uint8_t ser_txBuffReadPtr;
static volatile uint8_t ser_txBuffWritePtr;

void Ser_Init()
{
    // reset global vars
    ser_txBuffWritePtr = 0;
    ser_txBuffReadPtr = 0;
    
    UBRR1 = 1; // set baudrate to fastest
    
    // enable the UART, 9 bit, with RX interrupt always enabled
    UCSR1B |= _BV(RXEN1) | _BV(TXEN1) | _BV(RXCIE1) | _BV(UCSZ12);
    UCSR1C |= _BV(UCSZ11) | _BV(UCSZ10);
}

void Ser_Tx(uint16_t data)
{
    // push data to TX FIFO
    ser_txBuff[ser_txBuffWritePtr] = data;
    ser_txBuffWritePtr = (ser_txBuffWritePtr + 1) % SER_TXBUFFSIZE;
    UCSR1B |= _BV(UDRIE1); // this would initiate an interrupt immediately if nothing is currently transmitted
    // the interrupt is responsible for popping the FIFO
}

ISR(USART1_UDRE_vect)
{
    if (ser_txBuffWritePtr == ser_txBuffReadPtr)
    {
        // finished all transmissions, so disable the interrupt to stop
        UCSR1B &= ~_BV(UDRIE1);
    }
    else
    {
        uint16_t data = ser_txBuff[ser_txBuffReadPtr];
        
        // send the next data
        UCSR1B = ((data & 0x100) != 0) ? (UCSR1B | _BV(TXB81)) : (UCSR1B & ~_BV(TXB81));    
        UDR1 = data;
        
        // pop from FIFO
        ser_txBuffReadPtr = (ser_txBuffReadPtr + 1) % SER_TXBUFFSIZE;
    }
}