﻿using System;
using LibUsbDotNet;
using LibUsbDotNet.Main;
using System.Collections.Generic;
using System.Collections.Concurrent;
using System.Linq;
using System.Text;

namespace NehebkauWindowsHost
{
    public static class NehebkauDevice
    {
        private const int VID = 0x03EB;
        private const int PID = 0x2050;

        static LibUsbDotNet.DeviceNotify.WindowsDeviceNotifier wdn;
        static LibUsbDotNet.DeviceNotify.IDeviceNotifier dn;

        public static void StartNotifier()
        {
            wdn = new LibUsbDotNet.DeviceNotify.WindowsDeviceNotifier();
            wdn.OnDeviceNotify += new EventHandler<LibUsbDotNet.DeviceNotify.DeviceNotifyEventArgs>(wdn_OnDeviceNotify);
            wdn.Enabled = true;

            //dn = LibUsbDotNet.DeviceNotify.DeviceNotifier.OpenDeviceNotifier();
            //dn.OnDeviceNotify += new EventHandler<LibUsbDotNet.DeviceNotify.DeviceNotifyEventArgs>(dn_OnDeviceNotify);
            //dn.Enabled = true;

            if (LibUsbDotNet.UsbDevice.AllDevices.FirstOrDefault(i => i.Vid == VID && i.Pid == PID) != null)
            {
                if (OpenDevice())
                    frmMain.LogMsg("Master Device Opened");
            }
        }

        private static UsbDevice dev;
        private static UsbEndpointReader reader;
        private static UsbEndpointWriter writer;
        private static Queue<byte> receivedData = new Queue<byte>();

        public static bool IsConnected
        {
            get
            {
                return dev != null;
            }
        }

        public static bool IsOpen
        {
            get
            {
                if (dev == null)
                {
                    //if (OpenDevice())
                    //    frmMain.LogMsg("Master Device Opened");
                }

                if (dev == null)
                {
                    return false;
                }

                return dev.IsOpen;
            }
        }

        static void dn_OnDeviceNotify(object sender, LibUsbDotNet.DeviceNotify.DeviceNotifyEventArgs e)
        {
            wdn_OnDeviceNotify(sender, e);
        }

        static void wdn_OnDeviceNotify(object sender, LibUsbDotNet.DeviceNotify.DeviceNotifyEventArgs e)
        {
            if (e.Device.IdProduct != PID && e.Device.IdVendor != VID) return;

            frmMain.LogMsg(string.Format("USB Event: {0}", Enum.GetName(typeof(LibUsbDotNet.DeviceNotify.EventType), e.EventType)));

            if (e.EventType == LibUsbDotNet.DeviceNotify.EventType.DeviceArrival)
            {
                OpenDevice();
            }
            else if (e.EventType == LibUsbDotNet.DeviceNotify.EventType.CustomEvent)
            {
            }
            else if (e.EventType == LibUsbDotNet.DeviceNotify.EventType.DeviceTypeSpecific)
            {
            }
            else
            {
                CloseDevice();
            }
        }

        static void reader_DataReceived(object sender, EndpointDataEventArgs e)
        {
            frmMain.LogMsg(string.Format("DataReceived Event {0}", e.Count));
            byte[] buff = e.Buffer;
            int cnt = e.Count;
            ProcessIncoming(buff);
        }

        public static int TryReadControlTransfer()
        {
            if (IsOpen == false) return -1;

            try
            {
                byte[] buff = new byte[8];
                int txLen;
                UsbSetupPacket usp = new UsbSetupPacket(0x10 | (0x01 << 5) | 0x01, // dev to host, class, interface
                                                        0x01, // get report
                                                        0x00, // value
                                                        0x00, // index
                                                        0x08  // length
                                                        );
                if (dev.ControlTransfer(ref usp, buff, 8, out txLen))
                {
                    ProcessIncoming(buff);
                    return txLen;
                }
                else
                {
                    return -3;
                }
            }
            catch
            {
                return -2;
            }
        }

        public static int TryReader()
        {
            if (IsOpen == false) return -1;

            try
            {
                byte[] buff = new byte[8];
                int txLen;
                
                ErrorCode ec = reader.Read(buff, 0, 8, 100, out txLen);
                if (ec == ErrorCode.Ok)
                {
                    ProcessIncoming(buff);
                    return txLen;
                }
                else
                {
                    return -3;
                }
            }
            catch
            {
                return -2;
            }
        }

        public static int TryReadTransfer()
        {
            if (IsOpen == false) return -1;

            try
            {
                byte[] buff = new byte[8];
                int txLen;

                ErrorCode ec = reader.Transfer(buff, 0, 8, 100, out txLen);
                if (ec == ErrorCode.Ok)
                {
                    ProcessIncoming(buff);
                    return txLen;
                }
                else
                {
                    return -3;
                }
            }
            catch
            {
                return -2;
            }
        }

        private static void ProcessIncoming(byte[] buff)
        {
            try
            {
                bool slaveConnected = false;
                bool[] ledLocks = new bool[4];
                foreach (byte b in buff)
                {
                    if ((b & (1 << 6)) != 0)
                        receivedData.Enqueue(b);

                    if ((b & (1 << 6)) != 0)
                    {
                        slaveConnected = ((b & (1 << 5)) != 0);

                        for (int j = 0; j < 4; j++)
                            ledLocks[j] = ((b & (1 << j)) != 0);
                    }
                }
                frmMain.OnlyInstance.SetStatuses(slaveConnected, ledLocks);
            }
            catch (Exception ex)
            {
                CloseDevice();
                frmMain.LogMsg(string.Format("DataReceived Event Error: {0}", ex));
            }
        }

        public static void CloseDevice()
        {
            try { reader.DataReceivedEnabled = false; }
            catch { }
            try { reader.Dispose(); }
            catch { }
            try { writer.Dispose(); }
            catch { }
            try { dev.Close(); }
            catch { }
            try { wdn.Enabled = false; }
            catch { }
            try { dn.Enabled = false; }
            catch { }
            dev = null;
            reader = null;
            writer = null;
            wdn = null;
            dn = null;
        }

        private static bool OpenDevice()
        {
            try
            {
                TryOpenDevice();

                return true;
            }
            catch (Exception ex)
            {
                CloseDevice();
                frmMain.LogMsg(string.Format("Dev Open Err: {0}", ex));
                return false;
            }
        }

        public static void TryOpenDevice()
        {
            LibUsbDotNet.Main.UsbDeviceFinder udf = new LibUsbDotNet.Main.UsbDeviceFinder(VID, PID);

            UsbRegistry ur = LibUsbDotNet.UsbDevice.AllDevices.FirstOrDefault(i => i.Vid == VID && i.Pid == PID);
            if (ur != null)
            {
                if (ur.DeviceInterfaceGuids != null)
                {
                }
            }
            
            dev = UsbDevice.OpenUsbDevice(udf);
            dev.Open();
            reader = dev.OpenEndpointReader(ReadEndpointID.Ep01);
            reader.DataReceived += new EventHandler<EndpointDataEventArgs>(reader_DataReceived);
            reader.DataReceivedEnabled = true;
            writer = dev.OpenEndpointWriter(WriteEndpointID.Ep02);
        }

        public static int WriteKeyboard(byte modifier, byte[] keycodes)
        {
            if (IsOpen == false) return -2;

            try
            {
                byte[] data = new byte[keycodes.Length + 2];
                data[0] = 1; // indicate keyboard
                data[1] = modifier;
                for (int i = 0; i < keycodes.Length; i++)
                    data[2 + i] = keycodes[i];

                int txLen;
                dev.OpenEndpointWriter(WriteEndpointID.Ep02, EndpointType.Interrupt).Write(data, 0, 8, 100, out txLen);

                return txLen;
            }
            catch
            {
                CloseDevice();
                return -1;
            }
        }

        public static int WriteMouse(int button, int dX, int dY, int dW)
        {
            if (IsOpen == false) return -2;

            try
            {
                byte[] data = new byte[8];
                data[0] = 2; // indicate mouse
                data[1] = (byte)(button & 0xFF);
                data[2] = (byte)dX;
                data[3] = (byte)dY;
                data[4] = (byte)dW;

                int txLen;
                dev.OpenEndpointWriter(WriteEndpointID.Ep02, EndpointType.Interrupt).Write(data, 0, 8, 100, out txLen);

                return txLen;
            }
            catch
            {
                CloseDevice();
                return -1;
            }
        }
    }
}
